import pytest

from scanspec.core import Path
from scanspec.specs import Line


def test_line_path() -> None:
    x = "x"
    inst = Line(x, 0, 1, 5)
    dims = inst.calculate()
    path = Path(dims)
    assert len(path) == 5
    dim = path.consume()
    assert dim.midpoints == {x: pytest.approx([0, 0.25, 0.5, 0.75, 1.0])}
    assert len(path) == 0


def test_line_midpoints() -> None:
    x = "x"
    inst = Line(x, 0, 1, 5)
    it = inst.midpoints()
    assert it.axes == [x]
    assert len(it) == 5
    midpoints = [d[x] for d in it]
    assert midpoints == pytest.approx([0, 0.25, 0.5, 0.75, 1.0])



from typing import Any

import pytest

from scanspec.core import Path, SnakedFrames
from scanspec.regions import Circle, Ellipse, Polygon, Rectangle
from scanspec.specs import (
    DURATION,
    Concat,
    Line,
    Mask,
    Repeat,
    Spec,
    Spiral,
    Squash,
    Static,
    fly,
    step,
)

x, y, z = "x", "y", "z"


def ints(s):
    return pytest.approx([int(t) for t in s])


def test_one_point_duration() -> None:
    duration = Static.duration(1.0)
    (dim,) = duration.calculate()
    assert dim.midpoints == {DURATION: pytest.approx([1.0])}
    assert dim.lower == {DURATION: pytest.approx([1.0])}
    assert dim.upper == {DURATION: pytest.approx([1.0])}
    assert not isinstance(dim, SnakedFrames)
    assert dim.gap == ints("0")


def test_one_point_line() -> None:
    inst = Line(x, 0, 1, 1)
    (dim,) = inst.calculate()
    assert dim.midpoints == {x: pytest.approx([0])}
    assert dim.lower == {x: pytest.approx([-0.5])}
    assert dim.upper == {x: pytest.approx([0.5])}
    assert not isinstance(dim, SnakedFrames)
    assert dim.gap == ints("1")


def test_two_point_line() -> None:
    inst = Line(x, 0, 1, 2)
    (dim,) = inst.calculate()
    assert dim.midpoints == {x: pytest.approx([0, 1])}
    assert dim.lower == {x: pytest.approx([-0.5, 0.5])}
    assert dim.upper == {x: pytest.approx([0.5, 1.5])}
    assert dim.gap == ints("10")


def test_two_point_stepped_line() -> None:
    inst = step(Line(x, 0, 1, 2), 0.1)
    dimx, dimt = inst.calculate()
    assert dimx.midpoints == dimx.lower == dimx.upper == {x: pytest.approx([0, 1])}
    assert (
        dimt.midpoints == dimt.lower == dimt.upper == {DURATION: pytest.approx([0.1])}
    )
    assert inst.frames().gap == ints("11")


def test_two_point_fly_line() -> None:
    inst = fly(Line(x, 0, 1, 2), 0.1)
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1]),
        DURATION: pytest.approx([0.1, 0.1]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5]),
        DURATION: pytest.approx([0.1, 0.1]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5]),
        DURATION: pytest.approx([0.1, 0.1]),
    }
    assert dim.gap == ints("10")


def test_many_point_line() -> None:
    inst = Line(x, 0, 1, 5)
    (dim,) = inst.calculate()
    assert dim.midpoints == {x: pytest.approx([0, 0.25, 0.5, 0.75, 1])}
    assert dim.lower == {x: pytest.approx([-0.125, 0.125, 0.375, 0.625, 0.875])}
    assert dim.upper == {x: pytest.approx([0.125, 0.375, 0.625, 0.875, 1.125])}
    assert dim.gap == ints("10000")


def test_one_point_bounded_line() -> None:
    inst = Line.bounded(x, 0, 1, 1)
    assert inst == Line(x, 0.5, 1.5, 1)


def test_many_point_bounded_line() -> None:
    inst = Line.bounded(x, 0, 1, 4)
    assert inst == Line(x, 0.125, 0.875, 4)


def test_spiral() -> None:
    inst = Spiral(x, y, 0, 10, 5, 50, 10)
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        y: pytest.approx(
            [5.4, 6.4, 19.7, 23.8, 15.4, 1.7, -8.6, -10.7, -4.1, 8.3], abs=0.1
        ),
        x: pytest.approx(
            [0.3, -0.9, -0.7, 0.5, 1.5, 1.6, 0.7, -0.6, -1.8, -2.4], abs=0.1
        ),
    }
    assert dim.lower == {
        y: pytest.approx(
            [10.0, 2.7, 13.3, 23.5, 20.9, 8.7, -4.2, -10.8, -8.4, 1.6], abs=0.1
        ),
        x: pytest.approx(
            [0.0, -0.3, -1.0, -0.1, 1.1, 1.7, 1.3, 0.0, -1.2, -2.2], abs=0.1
        ),
    }
    assert dim.upper == {
        y: pytest.approx(
            [2.7, 13.3, 23.5, 20.9, 8.7, -4.2, -10.8, -8.4, 1.6, 15.3], abs=0.1
        ),
        x: pytest.approx(
            [-0.3, -1.0, -0.1, 1.1, 1.7, 1.3, 0.0, -1.2, -2.2, -2.4], abs=0.1
        ),
    }
    assert not isinstance(dim, SnakedFrames)
    assert dim.gap == ints("1000000000")


def test_spaced_spiral() -> None:
    inst = Spiral.spaced(x, y, 0, 10, 5, 1)
    assert inst == Spiral(x, y, 0, 10, 10, 10, 78)


def test_zipped_lines() -> None:
    inst = Line(x, 0, 1, 5).zip(Line(y, 1, 2, 5))
    assert inst.axes() == [x, y]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 0.25, 0.5, 0.75, 1]),
        y: pytest.approx([1, 1.25, 1.5, 1.75, 2]),
    }
    assert dim.gap == ints("10000")


def test_product_lines() -> None:
    inst = Line(y, 1, 2, 3) * Line(x, 0, 1, 2)
    assert inst.axes() == [y, x]
    dims = inst.calculate()
    assert len(dims) == 2
    dim = Path(dims).consume()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 0, 1]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.gap == ints("101010")


def test_zipped_product_lines() -> None:
    inst = Line(y, 1, 2, 3) * Line(x, 0, 1, 5).zip(Line(z, 2, 3, 5))
    assert inst.axes() == [y, x, z]
    dimy, dimxz = inst.calculate()
    assert dimxz.midpoints == {
        x: pytest.approx([0, 0.25, 0.5, 0.75, 1]),
        z: pytest.approx([2, 2.25, 2.5, 2.75, 3]),
    }
    assert dimy.midpoints == {
        y: pytest.approx([1, 1.5, 2]),
    }
    assert inst.frames().gap == ints("100001000010000")


def test_squashed_product() -> None:
    inst = Squash(Line(y, 1, 2, 3) * Line(x, 0, 1, 2))
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 0, 1]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.gap == ints("101010")


def test_squashed_multiplied_snake_scan() -> None:
    inst = Line(z, 1, 2, 2) * Squash(
        Line(y, 1, 2, 2) * ~Line.bounded(x, 3, 7, 2) * Static.duration(9, 2)
    )
    assert inst.axes() == [z, y, x, DURATION]
    dimz, dimxyt = inst.calculate()
    for d in dimxyt.midpoints, dimxyt.lower, dimxyt.upper:
        assert d == {
            x: pytest.approx([4, 4, 6, 6, 6, 6, 4, 4]),
            y: pytest.approx([1, 1, 1, 1, 2, 2, 2, 2]),
            DURATION: pytest.approx([9, 9, 9, 9, 9, 9, 9, 9]),
        }
    assert dimz.midpoints == dimz.lower == dimz.upper == {z: pytest.approx([1, 2])}
    assert inst.frames().gap == ints("1010101010101010")


def test_product_snaking_lines() -> None:
    inst = Line(y, 1, 2, 3) * ~Line(x, 0, 1, 2)
    assert inst.axes() == [y, x]
    dims = inst.calculate()
    assert len(dims) == 2
    dim = Path(dims).consume()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 1, 0, 0, 1]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5, 1.5, 0.5, -0.5, 0.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, -0.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.gap == ints("101010")


def test_concat_lines() -> None:
    inst = Concat(Line(x, 0, 1, 2), Line(x, 1, 2, 3))
    assert inst.axes() == [x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {x: pytest.approx([0, 1, 1, 1.5, 2])}
    assert dim.lower == {x: pytest.approx([-0.5, 0.5, 0.75, 1.25, 1.75])}
    assert dim.upper == {x: pytest.approx([0.5, 1.5, 1.25, 1.75, 2.25])}
    assert dim.gap == ints("10100")


def test_rect_region() -> None:
    inst = Line(y, 1, 3, 5) * Line(x, 0, 2, 3) & Rectangle(x, y, 0, 1, 1.5, 2.2)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 0, 1]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert dim.gap == ints("101010")


def test_rect_region_3D() -> None:
    inst = Static(z, 3.2, 2) * Line(y, 1, 3, 5) * Line(x, 0, 2, 3) & Rectangle(
        x, y, 0, 1, 1.5, 2.2
    )
    assert inst.axes() == [z, y, x]
    zdim, xydim = inst.calculate()
    assert zdim.midpoints == {z: pytest.approx([3.2, 3.2])}
    assert zdim.midpoints is zdim.upper
    assert zdim.midpoints is zdim.lower
    assert xydim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 0, 1]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert xydim.lower == {
        x: pytest.approx([-0.5, 0.5, -0.5, 0.5, -0.5, 0.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert xydim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2]),
    }
    assert inst.frames().gap == ints("101010101010")


def test_rect_region_union() -> None:
    inst = Line(y, 1, 3, 5) * Line(x, 0, 2, 3) & Rectangle(
        x, y, 0, 1, 1.5, 2.2
    ) | Rectangle(x, y, 0.5, 1.5, 2, 2.5)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 2, 0, 1, 2, 1, 2]),
        y: pytest.approx([1, 1, 1.5, 1.5, 1.5, 2, 2, 2, 2.5, 2.5]),
    }
    assert dim.gap == ints("1010010010")


def test_rect_region_intersection() -> None:
    inst = (
        Line(y, 1, 3, 5) * Line(x, 0, 2, 3)
        & Rectangle(x, y, 0, 1, 1.5, 2.2)
        & Rectangle(x, y, 0.5, 1.5, 2, 2.5)
    )
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([1, 1]),
        y: pytest.approx([1.5, 2]),
    }
    assert dim.gap == ints("11")


def test_rect_region_difference() -> None:
    # Bracket to force testing Mask.__sub__ rather than Region.__sub__
    inst = (
        Line(y, 1, 3, 5) * Line(x, 0, 2, 3).zip(Static(DURATION, 0.1))
        & Rectangle(x, y, 0, 1, 1.5, 2.2)
    ) - Rectangle(x, y, 0.5, 1.5, 2, 2.5)
    assert inst.axes() == [y, x, DURATION]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 0]),
        y: pytest.approx([1, 1, 1.5, 2]),
        DURATION: pytest.approx([0.1, 0.1, 0.1, 0.1]),
    }
    assert dim.gap == ints("1011")


def test_rect_region_symmetricdifference() -> None:
    inst = Line(y, 1, 3, 5) * Line(x, 0, 2, 3) & Rectangle(
        x, y, 0, 1, 1.5, 2.2
    ) ^ Rectangle(x, y, 0.5, 1.5, 2, 2.5)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 2, 0, 2, 1, 2]),
        y: pytest.approx([1, 1, 1.5, 1.5, 2, 2, 2.5, 2.5]),
    }
    assert dim.gap == ints("10111110")


def test_circle_region() -> None:
    inst = Line(y, 1, 3, 3) * Line(x, 0, 2, 3) & Circle(x, y, 1, 2, 1)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([1, 0, 1, 2, 1]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.lower == {
        x: pytest.approx([0.5, -0.5, 0.5, 1.5, 0.5]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.upper == {
        x: pytest.approx([1.5, 0.5, 1.5, 2.5, 1.5]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.gap == ints("11001")


def test_circle_snaked_region() -> None:
    inst = Mask(
        Line(y, 1, 3, 3) * ~Line(x, 0, 2, 3),
        Circle(x, y, 1, 2, 1),
        check_path_changes=False,
    )
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([1, 2, 1, 0, 1]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.lower == {
        x: pytest.approx([0.5, 2.5, 1.5, 0.5, 0.5]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.upper == {
        x: pytest.approx([1.5, 1.5, 0.5, -0.5, 1.5]),
        y: pytest.approx([1, 2, 2, 2, 3]),
    }
    assert dim.gap == ints("11001")


def test_ellipse_region() -> None:
    inst = Line("y", 1, 3, 3) * Line("x", 0, 2, 3) & Ellipse(x, y, 1, 2, 2, 1, 45)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([0, 1, 0, 1, 2, 1, 2]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3]),
    }
    assert dim.lower == {
        x: pytest.approx([-0.5, 0.5, -0.5, 0.5, 1.5, 0.5, 1.5]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3]),
    }
    assert dim.upper == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 2.5, 1.5, 2.5]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3]),
    }
    assert dim.gap == ints("1010010")


def test_polygon_region() -> None:
    x_verts = [0, 0.5, 4.0, 2.5]
    y_verts = [0, 3.5, 3.5, 0.5]
    inst = Line("y", 1, 3, 3) * Line("x", 0, 4, 5) & Polygon(x, y, x_verts, y_verts)
    assert inst.axes() == [y, x]
    (dim,) = inst.calculate()
    assert dim.midpoints == {
        x: pytest.approx([1, 2, 1, 2, 3, 1, 2, 3]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3, 3]),
    }
    assert dim.lower == {
        x: pytest.approx([0.5, 1.5, 0.5, 1.5, 2.5, 0.5, 1.5, 2.5]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3, 3]),
    }
    assert dim.upper == {
        x: pytest.approx([1.5, 2.5, 1.5, 2.5, 3.5, 1.5, 2.5, 3.5]),
        y: pytest.approx([1, 1, 2, 2, 2, 3, 3, 3]),
    }
    assert dim.gap == ints("10100100")


def test_xyz_stack() -> None:
    # Beam selector scan moves bounded between midpoints and lower and upper bounds at
    # maximum speed. Turnaround sections are where it sends the triggers
    spec = Line(z, 0, 1, 2) * ~Line(y, 0, 2, 3) * ~Line(x, 0, 3, 4)
    dim = spec.frames()
    assert len(dim) == 24
    assert dim.lower == {
        z: ints("000000000000111111111111"),
        y: ints("000011112222222211110000"),
        x: pytest.approx([-0.5, 0.5, 1.5, 2.5, 3.5, 2.5, 1.5, 0.5] * 3),
    }
    assert dim.upper == {
        z: ints("000000000000111111111111"),
        y: ints("000011112222222211110000"),
        x: pytest.approx([0.5, 1.5, 2.5, 3.5, 2.5, 1.5, 0.5, -0.5] * 3),
    }
    assert dim.midpoints == {
        z: ints("000000000000111111111111"),
        y: ints("000011112222222211110000"),
        x: ints("012332100123321001233210"),
    }
    assert dim.gap == ints("100010001000100010001000")
    # Check that it still works if you consume then start on a point that should
    # be False
    p = Path(spec.calculate())
    assert p.consume(4).gap == ints("1000")
    assert p.consume(4).gap == ints("1000")
    assert p.consume(5).gap == ints("10001")
    assert p.consume(2).gap == ints("00")
    assert p.consume().gap == ints("010001000")
    assert p.consume().gap == ints("")


def test_beam_selector() -> None:
    # Beam selector scan moves bounded between midpoints and lower and upper bounds at
    # maximum speed. Turnaround sections are where it sends the triggers
    spec = 10 * ~Line.bounded(x, 11, 19, 1)
    dim = spec.frames()
    assert len(dim) == 10
    assert dim.lower == {x: pytest.approx([11, 19, 11, 19, 11, 19, 11, 19, 11, 19])}
    assert dim.upper == {x: pytest.approx([19, 11, 19, 11, 19, 11, 19, 11, 19, 11])}
    assert dim.midpoints == {x: pytest.approx([15, 15, 15, 15, 15, 15, 15, 15, 15, 15])}
    assert dim.gap == ints("1111111111")


def test_gap_repeat() -> None:
    # Check that no gap propogates to dim.gap for snaked axis
    spec: Spec[Any] = Repeat(10, gap=False) * ~Line.bounded(x, 11, 19, 1)
    dim = spec.frames()
    assert len(dim) == 10
    assert dim.lower == {x: pytest.approx([11, 19, 11, 19, 11, 19, 11, 19, 11, 19])}
    assert dim.upper == {x: pytest.approx([19, 11, 19, 11, 19, 11, 19, 11, 19, 11])}
    assert dim.midpoints == {x: pytest.approx([15, 15, 15, 15, 15, 15, 15, 15, 15, 15])}
    assert dim.gap == ints("0000000000")


def test_gap_repeat_non_snake() -> None:
    # Check that no gap doesn't propogate to dim.gap for non-snaked axis
    spec: Spec[Any] = Repeat(3, gap=False) * Line.bounded(x, 11, 19, 1)
    dim = spec.frames()
    assert len(dim) == 3
    assert dim.lower == {x: pytest.approx([11, 11, 11])}
    assert dim.upper == {x: pytest.approx([19, 19, 19])}
    assert dim.midpoints == {x: pytest.approx([15, 15, 15])}
    assert dim.gap == ints("111")


def test_multiple_statics():
    part_1 = Static("y", 2) * Static("z", 3) * Line("x", 0, 10, 2)
    part_2 = Static("y", 4) * Static("z", 5) * Line("x", 0, 10, 2)
    spec = part_1.concat(part_2)
    assert list(spec.midpoints()) == [
        {"x": 0.0, "y": 2, "z": 3},
        {"x": 10.0, "y": 2, "z": 3},
        {"x": 0.0, "y": 4, "z": 5},
        {"x": 10.0, "y": 4, "z": 5},
    ]
    assert spec.frames().gap == ints("1010")


def test_multiple_statics_with_grid():
    grid = Line("y", 0, 10, 2) * Line("x", 0, 10, 2)
    part_1 = grid.zip(Static("a", 2)).zip(Static("b", 3))
    part_2 = grid.zip(Static("a", 4)).zip(Static("b", 5))
    spec = part_1.concat(part_2)
    assert list(spec.midpoints()) == [
        {"x": 0.0, "y": 0.0, "a": 2, "b": 3},
        {"x": 10.0, "y": 0.0, "a": 2, "b": 3},
        {"x": 0.0, "y": 10.0, "a": 2, "b": 3},
        {"x": 10.0, "y": 10.0, "a": 2, "b": 3},
        {"x": 0.0, "y": 0.0, "a": 4, "b": 5},
        {"x": 10.0, "y": 0.0, "a": 4, "b": 5},
        {"x": 0.0, "y": 10.0, "a": 4, "b": 5},
        {"x": 10.0, "y": 10.0, "a": 4, "b": 5},
    ]
    assert spec.frames().gap == ints("10101010")



"""
This file checks that all the example boilerplate text has been removed.
It can be deleted when all the contained tests pass
"""
from importlib.metadata import metadata
from pathlib import Path

ROOT = Path(__file__).parent.parent


def skeleton_check(check: bool, text: str):
    if ROOT.name == "python3-pip-skeleton" or str(ROOT) == "/project":
        # In the skeleton module the check should fail
        check = not check
        text = f"Skeleton didn't raise: {text}"
    if check:
        raise AssertionError(text)


def assert_not_contains_text(path: str, text: str, explanation: str):
    full_path = ROOT / path
    if full_path.exists():
        contents = full_path.read_text().replace("\n", " ")
        skeleton_check(text in contents, f"Please change ./{path} {explanation}")


# pyproject.toml
def test_module_summary():
    summary = metadata("scanspec")["summary"]
    skeleton_check(
        "One line description of your module" in summary,
        "Please change project.description in ./pyproject.toml "
        "to be a one line description of your module",
    )


# README
def test_changed_README_intro():
    assert_not_contains_text(
        "README.rst",
        "This is where you should write a short paragraph",
        "to include an intro on what your module does",
    )


def test_removed_adopt_skeleton():
    assert_not_contains_text(
        "README.rst",
        "This project contains template code only",
        "remove the note at the start",
    )


def test_changed_README_body():
    assert_not_contains_text(
        "README.rst",
        "This is where you should put some images or code snippets",
        "to include some features and why people should use it",
    )



import pytest

from scanspec.regions import Region
from scanspec.specs import Line, Spec, Squash


def test_not_implemented() -> None:
    with pytest.raises(NotImplementedError):
        Region().axis_sets()
    with pytest.raises(NotImplementedError):
        Region().mask({})
    with pytest.raises(NotImplementedError):
        Spec().axes()
    with pytest.raises(NotImplementedError):
        Spec().calculate()
    with pytest.raises(TypeError):
        Spec() * Region()


def test_non_snake_not_allowed_inside_snaking_dim() -> None:
    spec = Line("z", 1, 2, 2) * Squash(~Line("y", 1, 3, 3) * Line("x", 0, 2, 3))
    with pytest.raises(ValueError) as cm:
        spec.calculate()
    assert "['x'] would run backwards" in cm.value.args[0]


def test_snake_not_allowed_inside_odd_nested() -> None:
    spec = Line("z", 1, 2, 2) * Squash(Line("y", 1, 3, 3) * ~Line("x", 0, 2, 3))
    with pytest.raises(ValueError) as cm:
        spec.calculate()
    assert "['x'] would jump in position" in cm.value.args[0]



from typing import Any, Mapping

import pytest
from pydantic import ValidationError

from scanspec.regions import Circle, Rectangle, UnionOf
from scanspec.specs import Line, Mask, Spec, Spiral


def test_line_serializes() -> None:
    ob = Line("x", 0, 1, 4)
    serialized = {"type": "Line", "axis": "x", "start": 0.0, "stop": 1.0, "num": 4}
    assert ob.serialize() == serialized
    assert Spec.deserialize(serialized) == ob


def test_masked_circle_serializes() -> None:
    ob = Mask(Line("x", 0, 1, 4), Circle("x", "y", x_middle=0, y_middle=1, radius=4))
    serialized = {
        "type": "Mask",
        "spec": {"type": "Line", "axis": "x", "start": 0, "stop": 1, "num": 4},
        "region": {
            "x_axis": "x",
            "y_axis": "y",
            "x_middle": 0,
            "y_middle": 1,
            "radius": 4,
            "type": "Circle",
        },
        "check_path_changes": True,
    }
    assert ob.serialize() == serialized
    assert Spec.deserialize(serialized) == ob


def test_product_lines_serializes() -> None:
    ob = Line("z", 4, 5, 6) * Line("y", 2, 3, 5) * Line("x", 0, 1, 4)
    serialized = {
        "type": "Product",
        "outer": {
            "type": "Product",
            "outer": {
                "type": "Line",
                "axis": "z",
                "start": 4.0,
                "stop": 5.0,
                "num": 6,
            },
            "inner": {"type": "Line", "axis": "y", "start": 2.0, "stop": 3.0, "num": 5},
        },
        "inner": {"type": "Line", "axis": "x", "start": 0.0, "stop": 1.0, "num": 4},
    }

    assert ob.serialize() == serialized
    assert Spec.deserialize(serialized) == ob


def test_complex_nested_serializes() -> None:
    ob = Mask(
        Spiral.spaced("x", "y", 0, 0, 10, 3),
        UnionOf(
            Circle("x", "y", x_middle=0, y_middle=1, radius=4),
            Rectangle("x", "y", 0, 1.1, 1.5, 2.1, 30),
        ),
    )
    serialized = {
        "spec": {
            "x_axis": "x",
            "y_axis": "y",
            "x_start": 0.0,
            "y_start": 0.0,
            "x_range": 20.0,
            "y_range": 20.0,
            "num": 34,
            "rotate": 0.0,
            "type": "Spiral",
        },
        "region": {
            "left": {
                "x_axis": "x",
                "y_axis": "y",
                "x_middle": 0.0,
                "y_middle": 1.0,
                "radius": 4.0,
                "type": "Circle",
            },
            "right": {
                "x_axis": "x",
                "y_axis": "y",
                "x_min": 0.0,
                "y_min": 1.1,
                "x_max": 1.5,
                "y_max": 2.1,
                "angle": 30.0,
                "type": "Rectangle",
            },
            "type": "UnionOf",
        },
        "check_path_changes": True,
        "type": "Mask",
    }
    assert ob.serialize() == serialized
    assert Spec.deserialize(serialized) == ob


@pytest.mark.parametrize(
    "serialized",
    [
        {
            "axis": "x",
            "start": 0.0,
            "stop": 1.0,
            "num": 4,
            "foo": "bar",
            "type": "Line",
        },
        {
            "axis": "x",
            "start": 0.0,
            "stop": 1.0,
            "type": "Line",
        },
        {
            "axis": "x",
            "start": 0.0,
            "stop": {},
            "num": 4,
            "type": "Line",
        },
        {
            "axis": "x",
            "start": 0.0,
            "stop": None,
            "num": 4,
            "type": "Line",
        },
        {
            "type": "Product",
            "outer": None,
            "inner": {"type": "Line", "axis": "x", "start": 0.0, "stop": 1.0, "num": 4},
        },
    ],
    ids=["extra arg", "missing arg", "wrong type", "null value", "null spec"],
)
def test_detects_invalid_serialized(serialized: Mapping[str, Any]) -> None:
    with pytest.raises(ValidationError):
        Spec.deserialize(serialized)



import pathlib
import subprocess
import sys
from unittest.mock import patch

import matplotlib.pyplot as plt
import numpy as np
import pytest
from click.testing import CliRunner

from scanspec import __version__, cli
from scanspec.plot import _Arrow3D


def assert_min_max_2d(line, xmin, xmax, ymin, ymax, length=None):
    if length is not None:
        assert len(line.get_data()[0]) == length
    mins = np.min(line.get_data(), axis=1)
    maxs = np.max(line.get_data(), axis=1)
    assert list(zip(mins, maxs)) == [
        pytest.approx([xmin, xmax]),
        pytest.approx([ymin, ymax]),
    ]


def assert_min_max_3d(line, xmin, xmax, ymin, ymax, zmin, zmax, length=None):
    if length is not None:
        assert len(line.get_data_3d()[0]) == length
    mins = np.min(line.get_data_3d(), axis=1)
    maxs = np.max(line.get_data_3d(), axis=1)
    assert list(zip(mins, maxs)) == [
        pytest.approx([xmin, xmax]),
        pytest.approx([ymin, ymax]),
        pytest.approx([zmin, zmax]),
    ]


def assert_3d_arrow(artist, x, y, z):
    assert artist._verts3d[0][1] == pytest.approx(x)
    assert artist._verts3d[1][1] == pytest.approx(y)
    assert artist._verts3d[2][1] == pytest.approx(z)


def test_plot_1D_line() -> None:
    runner = CliRunner()
    spec = 'Line("x", 1, 2, 2)'
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.stdout == ""
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 4
    # Splines
    assert_min_max_2d(lines[0], 0.5, 1.5, 0, 0)
    assert_min_max_2d(lines[1], 1.5, 2.5, 0, 0)
    # Capture points
    assert_min_max_2d(lines[2], 1, 2, 0, 0, length=2)
    # End
    assert_min_max_2d(lines[3], 2.5, 2.5, 0, 0)
    # Arrows
    texts = axes.texts
    assert len(texts) == 1
    assert texts[0].xy == [0.5, 0]


def test_plot_1D_line_snake_repeat() -> None:
    runner = CliRunner()
    spec = '2 * ~Line.bounded("x", 1, 2, 1)'
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.stdout == ""
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 5
    # First repeat
    assert_min_max_2d(lines[0], 1, 2, 0, 0)
    # Turnaround
    assert_min_max_2d(lines[1], 2.0, 2.005585, 0, 0)
    # Second repeat
    assert_min_max_2d(lines[2], 1, 2, 0, 0)
    # Capture points
    assert_min_max_2d(lines[3], 1.5, 1.5, 0, 0, length=2)
    # End
    assert_min_max_2d(lines[4], 1, 1, 0, 0)
    # Arrows
    texts = axes.texts
    assert len(texts) == 2
    assert texts[0].xy == [1, 0]
    assert texts[1].xy == pytest.approx([2, 0])


def test_plot_1D_step() -> None:
    runner = CliRunner()
    spec = 'step(Line("x", 1, 2, 2), 0.1)'
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.stdout == ""
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 4
    # Start marker
    assert_min_max_2d(lines[0], 1, 1, 0, 0)
    # Step
    assert_min_max_2d(lines[1], 1, 2, 0, 0)
    # Capture points
    assert_min_max_2d(lines[2], 1, 2, 0, 0, length=2)
    # End
    assert_min_max_2d(lines[3], 2, 2, 0, 0)
    # Arrows
    texts = axes.texts
    assert len(texts) == 1
    assert texts[0].xy == [2, 0]


def test_plot_2D_line() -> None:
    runner = CliRunner()
    spec = 'Line("y", 2, 3, 2) * Snake(Line("x", 1, 2, 2))'
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.exit_code == 0
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 7
    # First row
    assert_min_max_2d(lines[0], 0.5, 1.5, 2, 2)
    assert_min_max_2d(lines[1], 1.5, 2.5, 2, 2)
    # Turnaround
    assert_min_max_2d(lines[2], 2.5, 2.7537562, 1.9999880, 3.0000119)
    # Second row
    assert_min_max_2d(lines[3], 1.5, 2.5, 3, 3)
    assert_min_max_2d(lines[4], 0.5, 1.5, 3, 3)
    # Capture points
    assert_min_max_2d(lines[5], 1, 2, 2, 3, length=4)
    # End
    assert_min_max_2d(lines[6], 0.5, 0.5, 3, 3)
    # Arrows
    texts = axes.texts
    assert len(texts) == 2
    assert texts[0].xy == [0.5, 2]
    assert texts[1].xy == pytest.approx([2.5, 3])


def test_plot_2D_line_rect_region() -> None:
    runner = CliRunner()
    spec = "Line(y, 1, 3, 5) * Line(x, 0, 2, 3) & Rectangle(x, y, 0, 1.1, 1.5, 2.1, 30)"
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.exit_code == 0
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 6
    # First row
    assert_min_max_2d(lines[0], -0.5, 0.5, 1.5, 1.5)
    # Turnaround
    assert_min_max_2d(lines[1], -0.6071045, 0.60710456, 1.4999969, 2.000003)
    # Second row
    assert_min_max_2d(lines[2], -0.5, 0.5, 2, 2)
    assert_min_max_2d(lines[3], 0.5, 1.5, 2, 2)
    # Capture points
    assert_min_max_2d(lines[4], 0, 1, 1.5, 2, length=3)
    # End
    assert_min_max_2d(lines[5], 1.5, 1.5, 2, 2)
    # Arrows
    texts = axes.texts
    assert len(texts) == 2
    assert texts[0].xy == [-0.5, 1.5]
    assert texts[1].xy == [-0.5, 2]
    # Regions
    patches = axes.patches
    assert len(patches) == 1
    assert type(patches[0]).__name__ == "Rectangle"
    assert patches[0].xy == (0, 1.1)
    assert patches[0].get_height() == 1.0
    assert patches[0].get_width() == 1.5
    assert patches[0].angle == 30


def test_plot_3D_line() -> None:
    runner = CliRunner()
    spec = 'Snake(Line("z", 5, 6, 2) * Line("y", 2, 3, 2) * Line("x", 1, 2, 2))'
    with patch("scanspec.plot.plt.show"):
        result = runner.invoke(cli.cli, ["plot", spec])
    assert result.exit_code == 0
    axes = plt.gcf().axes[0]
    lines = axes.lines
    assert len(lines) == 13
    # First grid
    # First row
    assert_min_max_3d(lines[0], 0.5, 1.5, 2, 2, 5, 5)
    assert_min_max_3d(lines[1], 1.5, 2.5, 2, 2, 5, 5)
    # Turnaround
    assert_min_max_3d(lines[2], 2.5, 2.7537562, 1.9999880, 3.0000119, 5, 5)
    # Second row
    assert_min_max_3d(lines[3], 1.5, 2.5, 3, 3, 5, 5)
    assert_min_max_3d(lines[4], 0.5, 1.5, 3, 3, 5, 5)
    # Turnaround
    assert_min_max_3d(lines[5], 0.2462437, 0.5, 3, 3, 4.9999880, 6.0000178)
    # Second grid
    # First row
    assert_min_max_3d(lines[6], 0.5, 1.5, 3, 3, 6, 6)
    assert_min_max_3d(lines[7], 1.5, 2.5, 3, 3, 6, 6)
    # Turnaround
    assert_min_max_3d(lines[8], 2.5, 2.7537562, 1.9999880, 3.0000119, 6, 6)
    # Second row
    assert_min_max_3d(lines[9], 1.5, 2.5, 2, 2, 6, 6)
    assert_min_max_3d(lines[10], 0.5, 1.5, 2, 2, 6, 6)
    # Capture points
    assert_min_max_3d(lines[11], 1, 2, 2, 3, 5, 6, length=8)
    # End
    assert_min_max_3d(lines[12], 0.5, 0.5, 2, 2, 6, 6)
    # Arrows
    extra_artists = axes.get_children()

    arrow_artists = list(
        filter(
            lambda artist: isinstance(artist, _Arrow3D)
            and artist.get_visible()
            and artist.get_in_layout(),
            extra_artists,
        )
    )
    assert len(arrow_artists) == 4
    assert_3d_arrow(arrow_artists[0], 0.5, 2, 5)
    assert_3d_arrow(arrow_artists[1], 2.5, 3, 5)
    assert_3d_arrow(arrow_artists[2], 0.5, 3, 6)
    assert_3d_arrow(arrow_artists[3], 2.5, 2, 6)


def test_schema() -> None:
    # If this test fails, regenerate the schema by running
    # scanspec schema > schema.json

    runner = CliRunner()
    result = runner.invoke(cli.schema)
    assert result.exit_code == 0
    schema_path = pathlib.Path(__file__).resolve().parent.parent / "schema.json"
    assert result.output == schema_path.read_text()


def test_cli_version():
    cmd = [sys.executable, "-m", "scanspec", "--version"]
    assert (
        subprocess.check_output(cmd).decode().strip()
        == f"scanspec, version {__version__}"
    )



from dataclasses import asdict
from typing import Any

import pytest
from fastapi.testclient import TestClient

from scanspec.service import PointsFormat, PointsRequest, app
from scanspec.specs import Line


@pytest.fixture
def client() -> TestClient:
    return TestClient(app)


# MIDPOINTS TEST(S) #
@pytest.mark.parametrize(
    "format,expected_midpoints",
    [
        (PointsFormat.FLOAT_LIST, [0.0, 0.25, 0.5, 0.75, 1.0]),
        (PointsFormat.STRING, "[0.   0.25 0.5  0.75 1.  ]"),
        (
            PointsFormat.BASE64_ENCODED,
            "AAAAAAAAAAAAAAAAAADQPwAAAAAAAOA/AAAAAAAA6D8AAAAAAADwPw==",
        ),
    ],
    ids=["float_list", "string", "base64"],
)
def test_midpoints(
    client: TestClient, format: PointsFormat, expected_midpoints: Any
) -> None:
    request = PointsRequest(Line("x", 0.0, 1.0, 5), max_frames=5, format=format)
    response = client.post("/midpoints", json=asdict(request))
    assert response.status_code == 200
    assert response.json() == {
        "total_frames": 5,
        "returned_frames": 5,
        "format": format.value,
        "midpoints": {"x": expected_midpoints},
    }


def test_subsampling(client: TestClient) -> None:
    spec = Line("x", 0, 10, 5) * Line("y", 0, 10, 5)
    request = PointsRequest(spec, max_frames=8, format=PointsFormat.FLOAT_LIST)
    response = client.post("/midpoints", json=asdict(request))
    assert response.status_code == 200
    assert response.json() == {
        "total_frames": 25,
        "returned_frames": 8,
        "format": "FLOAT_LIST",
        "midpoints": {"x": [0.0, 0.0, 10.0, 10.0], "y": [0.0, 10.0, 0.0, 10.0]},
    }


# BOUNDS TEST(S) #
@pytest.mark.parametrize(
    "format,expected_lower,expected_upper",
    [
        (
            PointsFormat.FLOAT_LIST,
            [-0.125, 0.125, 0.375, 0.625, 0.875],
            [0.125, 0.375, 0.625, 0.875, 1.125],
        ),
        (
            PointsFormat.STRING,
            "[-0.125  0.125  0.375  0.625  0.875]",
            "[0.125 0.375 0.625 0.875 1.125]",
        ),
        (
            PointsFormat.BASE64_ENCODED,
            "AAAAAAAAwL8AAAAAAADAPwAAAAAAANg/AAAAAAAA5D8AAAAAAADsPw==",
            "AAAAAAAAwD8AAAAAAADYPwAAAAAAAOQ/AAAAAAAA7D8AAAAAAADyPw==",
        ),
    ],
    ids=["float_list", "string", "base64"],
)
def test_bounds(
    client: TestClient, format: PointsFormat, expected_lower: Any, expected_upper: Any
) -> None:
    request = PointsRequest(Line("x", 0.0, 1.0, 5), max_frames=5, format=format)
    response = client.post("/bounds", json=asdict(request))
    assert response.status_code == 200
    assert response.json() == {
        "total_frames": 5,
        "returned_frames": 5,
        "format": format.value,
        "lower": {"x": expected_lower},
        "upper": {"x": expected_upper},
    }


# GAP TEST(S) #
def test_gap(client: TestClient) -> None:
    spec = Line("y", 0.0, 10.0, 3) * Line("x", 0.0, 10.0, 3)
    response = client.post("/gap", json=spec.serialize())
    assert response.status_code == 200
    assert response.json() == {
        "gap": [True, False, False, True, False, False, True, False, False]
    }


# SMALLEST STEP TEST(S) #
def test_smallest_step(client: TestClient) -> None:
    spec = Line("y", 0.0, 10.0, 3) * Line("x", 0.0, 10.0, 5)
    response = client.post("/smalleststep", json=spec.serialize())
    assert response.status_code == 200
    assert response.json() == {"absolute": 2.5, "per_axis": {"y": 5.0, "x": 2.5}}


# VALIDATE SPEC TEST(S) #
def test_validate_spec(client: TestClient) -> None:
    spec = Line.bounded("x", 0, 1, 5)
    response = client.post("/valid", json=spec.serialize())
    assert response.status_code == 200
    assert response.json() == {
        "input_spec": {
            "axis": "x",
            "start": 0.1,
            "stop": 0.9,
            "num": 5,
            "type": "Line",
        },
        "valid_spec": {
            "axis": "x",
            "start": 0.1,
            "stop": 0.9,
            "num": 5,
            "type": "Line",
        },
    }


def test_validate_invalid_spec(client: TestClient) -> None:
    spec = {"type": "Line", "axis": "x", "start": 0.0, "num": 10}
    response = client.post("/valid", json=spec)
    assert response.status_code == 422



import base64
import json
from enum import Enum
from typing import List, Mapping, Optional, Tuple, Union

import numpy as np
from fastapi import Body, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from pydantic import Field
from pydantic.dataclasses import dataclass

from scanspec.core import AxesPoints, Frames, Path

from .specs import Line, Spec

app = FastAPI()

#
# Data Model
#


#: A set of points, that can be returned in various formats
Points = Union[str, List[float]]


@dataclass
class ValidResponse:
    """Response model for spec validation."""

    input_spec: Spec = Field(description="The input scanspec")
    valid_spec: Spec = Field(description="The validated version of the spec")


class PointsFormat(str, Enum):
    """Formats in which we can return points."""

    STRING = "STRING"
    FLOAT_LIST = "FLOAT_LIST"
    BASE64_ENCODED = "BASE64_ENCODED"


@dataclass
class PointsRequest:
    """A request for generated scan points."""

    spec: Spec = Field(description="The spec from which to generate points")
    max_frames: Optional[int] = Field(
        description="The maximum number of points to return, if None will return "
        "as many as calculated",
        default=100000,
    )
    format: PointsFormat = Field(
        description="The format in which to output the points data",
        default=PointsFormat.FLOAT_LIST,
    )


@dataclass
class GeneratedPointsResponse:
    """Base class for responses that include generated point data."""

    total_frames: int = Field(description="Total number of frames in spec")
    returned_frames: int = Field(
        description="Total of number of frames in this response, may be "
        "less than total_frames due to downsampling etc."
    )
    format: PointsFormat = Field(description="Format of returned point data")


@dataclass
class MidpointsResponse(GeneratedPointsResponse):
    """Midpoints of a generated scan."""

    midpoints: Mapping[str, Points] = Field(
        description="The midpoints of scan frames for each axis"
    )


@dataclass
class BoundsResponse(GeneratedPointsResponse):
    """Bounds of a generated scan."""

    lower: Mapping[str, Points] = Field(
        description="Lower bounds of scan frames if different from midpoints"
    )
    upper: Mapping[str, Points] = Field(
        description="Upper bounds of scan frames if different from midpoints"
    )


@dataclass
class GapResponse:
    """Presence of gaps in a generated scan."""

    gap: List[bool] = Field(
        description="Boolean array indicating if there is a gap between each frame"
    )


@dataclass
class SmallestStepResponse:
    """Information about the smallest steps between points in a spec."""

    absolute: float = Field(
        description="Absolute smallest distance between two points on a single axis"
    )
    per_axis: Mapping[str, float] = Field(
        description="Smallest distance between two points on each axis"
    )


#
# API Routes
#

_EXAMPLE_SPEC = Line("y", 0.0, 10.0, 3) * Line("x", 0.0, 10.0, 4)
_EXAMPLE_POINTS_REQUEST = PointsRequest(
    _EXAMPLE_SPEC, max_frames=1024, format=PointsFormat.FLOAT_LIST
)


@app.post("/valid", response_model=ValidResponse)
def valid(
    spec: Spec = Body(..., example=_EXAMPLE_SPEC)
) -> Union[ValidResponse, JSONResponse]:
    """Validate wether a ScanSpec can produce a viable scan.

    Args:
        spec: The scanspec to validate

    Returns:
        ValidResponse: A canonical version of the spec if it is valid.
            An error otherwise.
    """
    valid_spec = Spec.deserialize(spec.serialize())
    return ValidResponse(spec, valid_spec)


@app.post("/midpoints", response_model=MidpointsResponse)
def midpoints(
    request: PointsRequest = Body(
        ...,
        example=_EXAMPLE_POINTS_REQUEST,
    )
) -> MidpointsResponse:
    """Generate midpoints from a scanspec.

    A scanspec can produce bounded points (i.e. a point is valid if an
    axis is between a minimum and and a maximum, see /bounds). The midpoints
    are the middle of each set of bounds.

    Args:
        request: Scanspec and formatting info.

    Returns:
        MidpointsResponse: Midpoints of the scan
    """
    chunk, total_frames = _to_chunk(request)
    return MidpointsResponse(
        total_frames,
        request.max_frames or total_frames,
        request.format,
        _format_axes_points(chunk.midpoints, request.format),
    )


@app.post("/bounds", response_model=BoundsResponse)
def bounds(
    request: PointsRequest = Body(
        ...,
        example=_EXAMPLE_POINTS_REQUEST,
    )
) -> BoundsResponse:
    """Generate bounds from a scanspec.

    A scanspec can produce points with lower and upper bounds.

    Args:
        request: Scanspec and formatting info.

    Returns:
        BoundsResponse: Bounds of the scan
    """
    chunk, total_frames = _to_chunk(request)
    return BoundsResponse(
        total_frames,
        request.max_frames or total_frames,
        request.format,
        _format_axes_points(chunk.lower, request.format),
        _format_axes_points(chunk.upper, request.format),
    )


@app.post("/gap", response_model=GapResponse)
def gap(
    spec: Spec = Body(
        ...,
        example=_EXAMPLE_SPEC,
    )
) -> GapResponse:
    """Generate gaps from a scanspec.

    A scanspec may indicate if there is a gap between two frames.
    The array returned corresponds to whether or not there is a gap
    after each frame.

    Args:
        request: Scanspec and formatting info.

    Returns:
        GapResponse: Bounds of the scan
    """
    dims = spec.calculate()  # Grab dimensions from spec
    path = Path(dims)  # Convert to a path
    gap = list(path.consume().gap)
    return GapResponse(gap)


@app.post("/smalleststep", response_model=SmallestStepResponse)
def smallest_step(
    spec: Spec = Body(..., example=_EXAMPLE_SPEC)
) -> SmallestStepResponse:
    """Calculate the smallest step in a scan, both absolutely and per-axis.

    Ignore any steps of size 0.

    Args:
        spec: The spec of the scan

    Returns:
        SmallestStepResponse: A description of the smallest steps in the spec
    """
    dims = spec.calculate()  # Grab dimensions from spec
    path = Path(dims)  # Convert to a path
    chunk = path.consume()

    absolute = _calc_smallest_step(list(chunk.midpoints.values()))
    per_axis = {
        axis: _calc_smallest_step([chunk.midpoints[axis]]) for axis in chunk.axes()
    }

    return SmallestStepResponse(absolute, per_axis)


#
# Utility Functions
#


def _to_chunk(request: PointsRequest) -> Tuple[Frames, int]:
    spec = Spec.deserialize(request.spec)
    dims = spec.calculate()  # Grab dimensions from spec
    path = Path(dims)  # Convert to a path

    # TOTAL FRAMES
    total_frames = len(path)  # Capture the total length of the path

    # MAX FRAMES
    # Limit the consumed data by the max_frames argument
    max_frames = request.max_frames
    if max_frames and (max_frames < len(path)):
        # Cap the frames by the max limit
        path = _reduce_frames(dims, max_frames)
    # WARNING: path object is consumed after this statement
    return path.consume(max_frames), total_frames


def _format_axes_points(
    axes_points: AxesPoints[str], format: PointsFormat
) -> Mapping[str, Points]:
    """Convert points to a requested format.

    Args:
        axes_points: The points to convert
        format: The target format

    Raises:
        KeyError: If the function does not support the given format

    Returns:
        Mapping[str, Points]: A mapping of axis to formatted points.
    """
    if format is PointsFormat.FLOAT_LIST:
        return {axis: list(points) for axis, points in axes_points.items()}
    elif format is PointsFormat.STRING:
        return {axis: str(points) for axis, points in axes_points.items()}
    elif format is PointsFormat.BASE64_ENCODED:
        return {
            axis: base64.b64encode(points.tobytes()).decode()
            for axis, points in axes_points.items()
        }
    else:
        raise KeyError(f"Unknown format: {format}")


def _reduce_frames(stack: List[Frames[str]], max_frames: int) -> Path:
    """Removes frames from a spec so len(path) < max_frames.

    Args:
        stack: A stack of Frames created by a spec
        max_frames: The maximum number of frames the user wishes to be returned
    """
    # Calculate the total number of frames
    num_frames = 1
    for frames in stack:
        num_frames *= len(frames)

    # Need each dim to be this much smaller
    ratio = 1 / np.power(max_frames / num_frames, 1 / len(stack))

    sub_frames = [_sub_sample(f, ratio) for f in stack]
    return Path(sub_frames)


def _sub_sample(frames: Frames[str], ratio: float) -> Frames:
    """Provides a sub-sample Frames object whilst preserving its core structure.

    Args:
        frames: the Frames object to be reduced
        ratio: the reduction ratio of the dimension
    """
    num_indexes = int(len(frames) / ratio)
    indexes = np.linspace(0, len(frames) - 1, num_indexes, dtype=np.int32)
    return frames.extract(indexes, calculate_gap=False)


def _calc_smallest_step(points: List[np.ndarray]) -> float:
    # Calc abs diffs of all axes, ignoring any zero values
    absolute_diffs = [_abs_diffs(axis_midpoints) for axis_midpoints in points]
    # Normalize and remove zeros
    norm_diffs = np.linalg.norm(absolute_diffs, axis=0)
    norm_diffs = norm_diffs[norm_diffs > 0.0]
    # Return the smallest value (Aka. smallest step)
    return np.amin(norm_diffs)


def _abs_diffs(array: np.ndarray) -> np.ndarray:
    """Calculates the absolute differences between adjacent elements in the array.

    Args:
        array: A 1xN array of numerical values

    Returns:
        A newly constucted array of absolute differences
    """
    # [array[1] - array[0], array[2] - array[1], ...]
    adjacent_diffs = array[1:] - array[:-1]
    return np.absolute(adjacent_diffs)


def run_app(cors: bool = False, port: int = 8080) -> None:
    """Run an application providing the scanspec service."""
    if cors:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

    import uvicorn

    uvicorn.run(app, port=port)


def scanspec_schema_text() -> str:
    """Generate the OpenAPI schema for the service as a string.

    Returns:
        str: The OpenAPI schema
    """
    return json.dumps(
        get_openapi(
            title=app.title,
            version=app.version,
            openapi_version=app.openapi_version,
            description=app.description,
            routes=app.routes,
        )
    )
